<!DOCTYPE html>
<html lang="en">
<head>
    <meta charset="UTF-8">
    <title>SPIRES: Results</title>
</head>
<body>
<div>
  <a href="/">Try another!</a>
</div>

{{ inner_html|safe }}

</body>
</html>
